# -*- coding: utf-8 -*-

"""This module provides access to the tenant REST api of Camunda."""

from __future__ import annotations
import dataclasses
import typing

import pycamunda
import pycamunda.base
import pycamunda.resource
from pycamunda.request import QueryParameter, PathParameter, BodyParameter

URL_SUFFIX = '/tenant'
URL_SUFFIX_USER_MEMBERS = '/user-members'
URL_SUFFIX_GROUP_MEMBERS = '/group-members'


__all__ = [
    'UserMemberCreate', 'UserMemberDelete', 'UserMemberOptions', 'GroupMemberCreate',
    'GroupMemberDelete', 'GroupMemberOptions', 'GetList', 'Count', 'Get', 'Create', 'Update',
    'Options', 'Delete'
]


@dataclasses.dataclass
class Tenant:
    """Data class of tenant as returned by the REST api of Camunda."""
    id_: str
    name: str

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> Tenant:
        return cls(id_=data['id'], name=data['name'])


class UserMemberCreate(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    user_id = PathParameter('userId')

    def __init__(self, url: str, id_: str, user_id: str):
        """Create a membership between a tenant and an user.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the tenant.
        :param user_id: Id of the user.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}' + URL_SUFFIX_USER_MEMBERS + '/{userId}')
        self.id_ = id_
        self.user_id = user_id

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        super().__call__(pycamunda.base.RequestMethod.PUT, *args, **kwargs)


class UserMemberDelete(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    user_id = PathParameter('userId')

    def __init__(self, url: str, id_: str, user_id: str):
        """Delete a membership between a tenant and an user.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the tenant.
        :param user_id: Id of the user.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}' + URL_SUFFIX_USER_MEMBERS + '/{userId}')
        self.id_ = id_
        self.user_id = user_id

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        super().__call__(pycamunda.base.RequestMethod.DELETE, *args, **kwargs)


class UserMemberOptions(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')

    def __init__(self, url: str, id_: str):
        """Get a list of options the currently authenticated user can perform on the tenant
        membership resource.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the tenant.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}' + URL_SUFFIX_USER_MEMBERS)
        self.id_ = id_

    def __call__(self, *args, **kwargs) -> pycamunda.resource.ResourceOptions:
        """Send the request"""
        response = super().__call__(pycamunda.base.RequestMethod.OPTIONS, *args, **kwargs)

        return pycamunda.resource.ResourceOptions.load(response.json())


class GroupMemberCreate(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    group_id = PathParameter('groupId')

    def __init__(self, url: str, id_: str, group_id: str):
        """Create a membership between a tenant and a group.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the tenant.
        :param group_id: Id of the group.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}' + URL_SUFFIX_GROUP_MEMBERS + '/{groupId}')
        self.id_ = id_
        self.group_id = group_id

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        super().__call__(pycamunda.base.RequestMethod.PUT, *args, **kwargs)


class GroupMemberDelete(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    group_id = PathParameter('groupId')

    def __init__(self, url: str, id_: str, group_id: str):
        """Delete a membership between a tenant and a group.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the tenant.
        :param group_id: Id of the group.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}' + URL_SUFFIX_GROUP_MEMBERS + '/{groupId}')
        self.id_ = id_
        self.group_id = group_id

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        super().__call__(pycamunda.base.RequestMethod.DELETE, *args, **kwargs)


class GroupMemberOptions(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')

    def __init__(self, url: str, id_: str):
        """Get a list of options the currently authenticated user can perform on the tenant
        membership resource.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the tenant.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}' + URL_SUFFIX_GROUP_MEMBERS)
        self.id_ = id_

    def __call__(self, *args, **kwargs) -> pycamunda.resource.ResourceOptions:
        """Send the request"""
        response = super().__call__(pycamunda.base.RequestMethod.OPTIONS, *args, **kwargs)

        return pycamunda.resource.ResourceOptions.load(response.json())


class GetList(pycamunda.base.CamundaRequest):

    id_ = QueryParameter('id')
    name = QueryParameter('name')
    name_like = QueryParameter('nameLike')
    user_member = QueryParameter('userMember')
    group_member = QueryParameter('groupMember')
    including_groups_of_user = QueryParameter(
        'includingGroupsOfUser',
        provide=pycamunda.base.value_is_true
    )
    sort_by = QueryParameter('sortBy', mapping={'id_': 'id', 'name': 'name'})
    ascending = QueryParameter(
        'sortOrder',
        mapping={True: 'asc', False: 'desc'},
        provide=lambda self, obj, obj_type: vars(obj).get('sort_by', None) is not None
    )
    first_result = QueryParameter('firstResult')
    max_results = QueryParameter('maxResults')

    def __init__(
        self,
        url: str,
        id_: str = None,
        name: str = None,
        name_like: str = None,
        user_member: str = None,
        group_member: str = None,
        including_groups_of_user: bool = False,
        sort_by: str = None,
        ascending: bool = True,
        first_result: int = None,
        max_results: int = None
    ):
        """Query for a list of tenants using a list of parameters. The size of the result set can be
        retrieved by using the Count class.

        :param url: Camunda Rest engine URL.
        :param id_: Filter by the id of the tenant.
        :param name: Filter by the name of the tenant.
        :param name_like: Filter by a substring of the name.
        :param user_member: Filter for tenants where the given user is member of.
        :param group_member: Filter for tenants where  the given group is a member of.
        :param including_groups_of_user: Filter for tenants where the user or one of his groups is a
                                         member of. Can only be used with `user_member` parameter.
        :param sort_by: Sort the results by `id_` or `name` of the tenant.
        :param ascending: Sort order.
        :param first_result: Pagination of results. Index of the first result to return.
        :param max_results: Pagination of results. Maximum number of results to return.
        """
        super().__init__(url=url + URL_SUFFIX)
        self.id_ = id_
        self.name = name
        self.name_like = name_like
        self.user_member = user_member
        self.group_member = group_member
        self.including_groups_of_user = including_groups_of_user
        self.sort_by = sort_by
        self.ascending = ascending
        self.first_result = first_result
        self.max_results = max_results

    def __call__(self, *args, **kwargs) -> typing.Tuple[Tenant]:
        """Send the request"""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return tuple(Tenant.load(tenant_json) for tenant_json in response.json())


class Count(pycamunda.base.CamundaRequest):

    id_ = QueryParameter('id')
    name = QueryParameter('name')
    name_like = QueryParameter('nameLike')
    user_member = QueryParameter('userMember')
    group_member = QueryParameter('groupMember')
    including_groups_of_user = QueryParameter(
        'includingGroupsOfUser',
        provide=pycamunda.base.value_is_true
    )

    def __init__(
        self,
        url: str,
        id_: str = None,
        name: str = None,
        name_like: str = None,
        user_member: str = None,
        group_member: str = None,
        including_groups_of_user: bool = False
    ):
        """Count tenants.

        :param url: Camunda Rest engine URL.
        :param id_: Filter by the id of the tenant.
        :param name: Filter by the name of the tenant.
        :param name_like: Filter by a substring of the name.
        :param user_member: Filter for tenants where the given user is member of.
        :param group_member: Filter for tenants where  the given group is a member of.
        :param including_groups_of_user: Filter for tenants where the user or one of his groups is a
                                         member of. Can only be used with `user_member` parameter.
        """
        super().__init__(url=url + URL_SUFFIX + '/count')
        self.id_ = id_
        self.name = name
        self.name_like = name_like
        self.user_member = user_member
        self.group_member = group_member
        self.including_groups_of_user = including_groups_of_user

    def __call__(self, *args, **kwargs) -> int:
        """Send the request"""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return response.json()['count']


class Get(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')

    def __init__(self, url: str, id_: str):
        """Get a tenant.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the tenant.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}')
        self.id_ = id_

    def __call__(self, *args, **kwargs) -> Tenant:
        """Send the request"""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return Tenant.load(response.json())


class Create(pycamunda.base.CamundaRequest):

    id_ = BodyParameter('id')
    name = BodyParameter('name')

    def __init__(self, url: str, id_: str, name: str):
        """Create a new tenant.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the tenant.
        :param name: Name of the tenant.
        """
        super().__init__(url=url + URL_SUFFIX + '/create')
        self.id_ = id_
        self.name = name

    def __call__(self, *args, **kwargs) -> None:
        """Send the request"""
        super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)


class Update(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    new_id = BodyParameter('id')
    new_name = BodyParameter('name')

    def __init__(self, url: str, id_: str, new_id: str, new_name: str):
        """Update a tenant.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the tenant.
        :param new_id: New id of the tenant.
        :param new_name: New name of the tenant.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}')
        self.id_ = id_
        self.new_id = new_id
        self.new_name = new_name

    def __call__(self, *args, **kwargs) -> None:
        """Send the request"""
        super().__call__(pycamunda.base.RequestMethod.PUT, *args, **kwargs)


class Options(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')

    def __init__(self, url: str, id_: str = None):
        """Get a list of options the currently authenticated user can perform on the tenant
        resource.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the tenant.
        """
        super().__init__(url=url + URL_SUFFIX)
        self.id_ = id_

    @property
    def url(self):
        return super().url + (f'/{self.id_}' if self.id_ is not None else '')

    def __call__(self, *args, **kwargs) -> pycamunda.resource.ResourceOptions:
        """Send the request"""
        response = super().__call__(pycamunda.base.RequestMethod.OPTIONS, *args, **kwargs)

        return pycamunda.resource.ResourceOptions.load(response.json())


class Delete(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')

    def __init__(self, url: str, id_: str):
        """Delete a tenant.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the tenant.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}')
        self.id_ = id_

    def __call__(self, *args, **kwargs) -> None:
        """Send the request"""
        super().__call__(pycamunda.base.RequestMethod.DELETE, *args, **kwargs)



# -*- coding: utf-8 -*-

"""This module provides access to the identity REST api of Camunda."""

from __future__ import annotations
import dataclasses
import typing

import pycamunda
import pycamunda.base
import pycamunda.group
import pycamunda.user
from pycamunda.request import QueryParameter, PathParameter, BodyParameter, BodyParameterContainer

URL_SUFFIX = '/identity'


__all__ = ['GetGroups', 'VerifyUser', 'GetPasswordPolicy', 'ValidatePassword']


@dataclasses.dataclass
class UsersGroups:
    """Data class of the groups of an user and all of their members as returned by the REST api of
    Camunda."""
    groups: typing.Tuple[pycamunda.group.Group]
    group_users: typing.Tuple[pycamunda.user.User]

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> UsersGroups:
        return cls(
            groups=tuple(pycamunda.group.Group.load(group) for group in data['groups']),
            group_users=tuple(pycamunda.user.User.load(user) for user in data['groupUsers'])
        )


@dataclasses.dataclass
class AuthStatus:
    """Data class of the authentication status of an user as returned by the REST api of Camunda."""
    user_id: str
    authenticated: bool

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> AuthStatus:
        return cls(
            user_id=data['authenticatedUser'],
            authenticated=data['authenticated']
        )


@dataclasses.dataclass
class PasswordPolicy:
    """Data class of the password policy as returned by the REST api of Camunda."""
    placeholder: str
    parameters: typing.Dict[str, typing.Any]

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> PasswordPolicy:
        return cls(
            placeholder=data['placeholder'],
            parameters=data['parameters']
        )


@dataclasses.dataclass
class PasswordPolicyCompliance:
    """Data class of the password policy compliance as returned by the REST api of Camunda."""
    policy: PasswordPolicy
    valid: bool

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> PasswordPolicyCompliance:
        policy = PasswordPolicy.load(
            {k: v for k, v in data.items() if k in ('placeholder', 'parameters')}
        )
        return cls(policy=policy, valid=data['valid'])


class GetGroups(pycamunda.base.CamundaRequest):

    user_id = QueryParameter('userId')

    def __init__(self, url: str, user_id: str):
        """Get the groups of an user and all of their members.

        :param url: Camunda Rest engine URL.
        :param user_id: Id of the user.
        """
        super().__init__(url=url + URL_SUFFIX + '/groups')
        self.user_id = user_id

    def __call__(self, *args, **kwargs) -> UsersGroups:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return UsersGroups.load(response.json())


class VerifyUser(pycamunda.base.CamundaRequest):

    username = BodyParameter('username')
    password = BodyParameter('password')

    def __init__(self, url: str, username: str, password: str):
        """Verify the credentials of an user.

        :param url: Camunda Rest engine URL.
        :param username: Name of the user.
        :param password: Password of the user.
        """
        super().__init__(url=url + URL_SUFFIX + '/verify')
        self.username = username
        self.password = password

    def __call__(self, *args, **kwargs) -> AuthStatus:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)

        return AuthStatus.load(response.json())


class GetPasswordPolicy(pycamunda.base.CamundaRequest):

    def __init__(self, url: str):
        """Get the list of password policy rules.

        :param url: Camunda Rest engine URL.
        """
        super().__init__(url=url + URL_SUFFIX + '/password-policy')

    def __call__(self, *args, **kwargs) -> typing.Tuple[PasswordPolicy]:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return tuple(PasswordPolicy.load(policy_json) for policy_json in response.json()['rules'])


class ValidatePassword(pycamunda.base.CamundaRequest):

    password = BodyParameter('password')

    def __init__(self, url: str, password: str):
        """Validate a password against the password policy rules.

        :param url: Camunda Rest engine URL.
        :param password: Password to validate.
        """
        super().__init__(url=url + URL_SUFFIX + '/password-policy')
        self.password = password

    def __call__(self, *args, **kwargs) -> typing.Tuple[PasswordPolicyCompliance]:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)

        return tuple(
            PasswordPolicyCompliance.load(policy_comp_json)
            for policy_comp_json in response.json()['rules']
        )



# -*- coding: utf-8 -*-

"""This module provides access to the execution REST api of Camunda."""

from __future__ import annotations
import dataclasses
import typing


__all__ = []


@dataclasses.dataclass
class Execution:
    """Data class of execution as returned by the REST api of Camunda."""
    id_: str
    process_instance_id: str
    ended: bool
    tenant_id: str

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> Execution:
        return cls(
            id_=data['id'],
            process_instance_id=data['processInstanceId'],
            ended=data['ended'],
            tenant_id=data['tenantId']
        )



# -*- coding: utf-8 -*-

"""This module provides access to the process instance REST api of Camunda."""

from __future__ import annotations
import dataclasses
import typing

import pycamunda.resource
import pycamunda.variable
import pycamunda.activityinst
import pycamunda.instruction
import pycamunda.incident
import pycamunda.batch
import pycamunda.base
from pycamunda.request import QueryParameter, PathParameter, BodyParameter

URL_SUFFIX = '/process-instance'


__all__ = ['Delete', 'GetActivityInstance', 'GetList', 'Get', 'Modify', 'Activate', 'Suspend']


@dataclasses.dataclass
class ProcessInstance:
    """Data class of process instance as returned by the REST api of Camunda."""
    id_: str
    definition_id: str
    business_key: str
    case_instance_id: str
    tenant_id: str
    suspended: bool
    links: typing.Tuple[pycamunda.resource.Link]
    variables: typing.Dict[str, pycamunda.variable.Variable] = None

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> ProcessInstance:
        process_instance = cls(
            id_=data['id'],
            definition_id=data['definitionId'],
            business_key=data['businessKey'],
            case_instance_id=data['caseInstanceId'],
            tenant_id=data['tenantId'],
            suspended=data['suspended'],
            links=tuple(pycamunda.resource.Link.load(link_json) for link_json in data['links']),
        )
        try:
            variables = data['variables']
        except KeyError:
            pass
        else:
            process_instance.variables = {name: pycamunda.variable.Variable.load(var_json)
                                          for name, var_json in variables.items()}

        return process_instance


class Delete(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    skip_custom_listeners = QueryParameter('skipCustomListeners')
    skip_io_mappings = QueryParameter('skipIoMappings')
    skip_subprocesses = QueryParameter('skipSubprocesses')
    fail_if_not_exists = QueryParameter('failIfNotExists')

    def __init__(
            self,
            url: str,
            id_: str,
            skip_custom_listeners: bool = False,
            skip_io_mappings: bool = False,
            skip_subprocesses: bool = False,
            fail_if_not_exists: bool = True
    ):
        """Delete a process instance.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the process instance.
        :param skip_custom_listeners: Whether to skip custom listeners and notify only builtin ones.
        :param skip_io_mappings: Whether to skip input/output mappings.
        :param skip_subprocesses: Whether to skip subprocesses.
        :param fail_if_not_exists: Whether to fail if the provided process instance id is not found.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}')
        self.id_ = id_
        self.skip_custom_listeners = skip_custom_listeners
        self.skip_io_mappings = skip_io_mappings
        self.skip_subprocesses = skip_subprocesses
        self.fail_if_not_exists = fail_if_not_exists

    def __call__(self, *args, **kwargs):
        """Send the request."""
        super().__call__(pycamunda.base.RequestMethod.DELETE, *args, **kwargs)


class GetActivityInstance(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')

    def __init__(self, url: str, id_: str):
        """Get an activity instance tree for a specific process instance.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the process instance.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/activity-instances')
        self.id_ = id_

    def __call__(self, *args, **kwargs) -> pycamunda.activityinst.ActivityInstance:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return pycamunda.activityinst.ActivityInstance.load(response.json())


class GetList(pycamunda.base.CamundaRequest):

    process_instance_ids = QueryParameter('processInstanceIds')
    business_key = QueryParameter('businessKey')
    business_key_like = QueryParameter('businessKeyLike')
    case_instance_id = QueryParameter('caseInstanceId')
    process_definition_id = QueryParameter('processDefinitionId')
    process_definition_key = QueryParameter('processDefinitionKey')
    process_definition_key_in = QueryParameter('processDefinitionKeyIn')
    process_definition_key_not_in = QueryParameter('processDefinitionKeyNotIn')
    deployment_id = QueryParameter('deploymentId')
    super_process_instance = QueryParameter('superProcessInstance')
    sub_process_instance = QueryParameter('subProcessInstance')
    super_case_instance_id = QueryParameter('superCaseInstance')
    sub_case_instance_id = QueryParameter('subCaseInstance')
    active = QueryParameter('active', provide=pycamunda.base.value_is_true)
    suspended = QueryParameter('suspended', provide=pycamunda.base.value_is_true)
    with_incident = QueryParameter('withIncident')
    incident_id = QueryParameter('incidentId')
    incident_type = QueryParameter('incidentType')
    incident_message = QueryParameter('incidentMessage')
    incident_message_like = QueryParameter('incidentMessageLike')
    tenant_id_in = QueryParameter('tenantIdIn')
    without_tenant_id = QueryParameter('withoutTenantId', provide=pycamunda.base.value_is_true)
    activity_id_in = QueryParameter('activityIdIn')
    root_process_instances = QueryParameter('rootProcessInstances')
    leaf_process_instances = QueryParameter('leafProcessInstances')
    process_definition_without_tenant_id_in = QueryParameter('processDefinitionWithoutTenantIdIn')
    variables = QueryParameter('variables')
    variable_names_ignore_case = QueryParameter('variableNamesIgnoreCase')
    variable_values_ignore_case = QueryParameter('variableValuesIgnoreCase')
    sort_by = QueryParameter(
        'sortBy',
        mapping={
            'instance_id': 'instanceId',
            'definition_key': 'definitionKey',
            'definition_id': 'definitionId',
            'tenant_id': 'tenantId',
            'business_key': 'businessKey'
        }
    )
    ascending = QueryParameter(
        'sortOrder',
        mapping={True: 'asc', False: 'desc'},
        provide=lambda self, obj, obj_type: vars(obj).get('sort_by', None) is not None
    )
    first_result = QueryParameter('firstResult')
    max_results = QueryParameter('maxResults')

    def __init__(
            self,
            url: str,
            process_instance_ids: typing.Iterable[str] = None,
            business_key: str = None,
            business_key_like: str = None,
            case_instance_id: str = None,
            process_definition_id: str = None,
            process_definition_key: str = None,
            process_definition_key_in: typing.Iterable[str] = None,
            process_definition_key_not_in: typing.Iterable[str] = None,
            deployment_id: str = None,
            super_process_instance: str = None,
            sub_process_instance: str = None,
            super_case_instance_id: str = None,
            sub_case_instance_id: str = None,
            active: bool = False,
            suspended: bool = False,
            with_incident: bool = None,
            incident_id: str = None,
            incident_type: typing.Union[str, pycamunda.incident.IncidentType] = None,
            incident_message: str = None,
            incident_message_like: str = None,
            tenant_id_in: typing.Iterable[str] = None,
            without_tenant_id: bool = False,
            activity_id_in: typing.Iterable[str] = None,
            root_process_instances: bool = None,
            leaf_process_instances: bool = None,
            process_definition_without_tenant_id_in: bool = None,
            variables=None,  # TODO add annotation
            variable_names_ignore_case: bool = None,
            variable_values_ignore_case: bool = None,
            sort_by: str = None,
            ascending: bool = True,
            first_result: int = None,
            max_results: int = None
    ):
        """Get a list of process instances.

        :param url: Camunda Rest engine URL.
        :param process_instance_ids: Filter by process instance ids.
        :param business_key: Filter by business key.
        :param business_key_like: Filter by a substring of the business key.
        :param case_instance_id: Filter by case instance id.
        :param process_definition_id: Filter by process definition id.
        :param process_definition_key: Filter by process definition key.
        :param process_definition_key_in: Filter whether the process definition key is one of
                                          multiple ones.
        :param process_definition_key_not_in: Filter whether the process definition key is not one
                                              of multiple ones.
        :param deployment_id: Filter by deployment id.
        :param super_process_instance: Filter process instances that are a sub process of the
                                          provided id.
        :param sub_process_instance: Filter process instances that are a super process of the
                                        provided id.
        :param super_case_instance_id: Filter process instances that are a sub process of the
                                       provided case instance id.
        :param sub_case_instance_id: Filter process instances that are a super process of the
                                     provided case instance id.
        :param active: Whether to include only active process instances.
        :param suspended: Whether to include only suspended process instances.
        :param with_incident: Whether to include only process instances that have incidents.
        :param incident_id: Filter by incident id.
        :param incident_type: Filter by incident type.
        :param incident_message: Filter by the incident message.
        :param incident_message_like: Filter by a substring the incident message.
        :param tenant_id_in: Filter whether the tenant id is one of multiple ones.
        :param without_tenant_id: Whether to include only process instances that belong to no
                                  tenant.
        :param activity_id_in: Filter whether the activity id is one of multiple ones.
        :param root_process_instances: Include only top level process instances.
        :param leaf_process_instances: Include only bottom level process instances.
        :param process_definition_without_tenant_id_in: Include only process instance where the
                                                     process definition has no tenant id.
        :param variables: # TODO (add via its own method?)
        :param variable_names_ignore_case: Whether to ignore case sensitivity for variables names.
        :param variable_values_ignore_case: Whether to ignore case sensitivity for variable values.
        :param sort_by: Sort the results by 'instance_id', 'definition_key', 'definition_id',
                        'tenant_id', 'business_key'.
        :param ascending: Sort order.
        :param first_result: Pagination of results. Index of the first result to return.
        :param max_results: Pagination of results. Maximum number of results to return.
        """
        super().__init__(url=url + URL_SUFFIX)
        self.process_instance_ids = process_instance_ids
        self.business_key = business_key
        self.business_key_like = business_key_like
        self.case_instance_id = case_instance_id
        self.process_definition_id = process_definition_id
        self.process_definition_key = process_definition_key
        self.process_definition_key_in = process_definition_key_in
        self.process_definition_key_not_in = process_definition_key_not_in
        self.deployment_id = deployment_id
        self.super_process_instance = super_process_instance
        self.sub_process_instance = sub_process_instance
        self.super_case_instance_id = super_case_instance_id
        self.sub_case_instance_id = sub_case_instance_id
        self.active = active
        self.suspended = suspended
        self.with_incident = with_incident
        self.incident_id = incident_id
        self.incident_type = None
        if incident_type is not None:
            self.incident_type = pycamunda.incident.IncidentType(incident_type)
        self.incident_message = incident_message
        self.incident_message_like = incident_message_like
        self.tenant_id_in = tenant_id_in
        self.without_tenant_id = without_tenant_id
        self.activity_id_in = activity_id_in
        self.root_process_instances = root_process_instances
        self.leaf_process_instances = leaf_process_instances
        self.process_definition_without_tenant_id_in = process_definition_without_tenant_id_in
        self.variables = variables
        self.variable_names_ignore_case = variable_names_ignore_case
        self.variable_values_ignore_case = variable_values_ignore_case
        self.sort_by = sort_by
        self.ascending = ascending
        self.first_result = first_result
        self.max_results = max_results

    def __call__(self, *args, **kwargs) -> typing.Tuple[ProcessInstance]:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return tuple(ProcessInstance.load(instance_json) for instance_json in response.json())


class Get(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')

    def __init__(self, url: str, id_: str):
        """Get a process instance.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the process instance.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}')
        self.id_ = id_

    def __call__(self, *args, **kwargs) -> ProcessInstance:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return ProcessInstance.load(response.json())


class Modify(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    skip_custom_listeners = BodyParameter('skipCustomListeners')
    skip_io_mappings = BodyParameter('skipIoMappings')
    instructions = BodyParameter('instructions')
    annotation = BodyParameter('annotation')

    def __init__(
        self,
        url: str,
        id_: str,
        async_: bool = False,
        skip_custom_listeners: bool = False,
        skip_io_mappings: bool = False,
        annotation: str = None
    ):
        """Modify a running process instance. This is done by adding instructions that are executed
        on a process instance.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the process instance.
        :param async_: Whether to run this request asynchronously.
        :param skip_custom_listeners: Whether to skip custom listeners and notify only builtin ones.
        :param skip_io_mappings: Whether to skip input/output mappings.
        :param annotation: Arbitrary text annotation for auditing reasons.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/modification')
        self.id_ = id_
        self.async_ = async_
        self.skip_custom_listeners = skip_custom_listeners
        self.skip_io_mappings = skip_io_mappings
        self.annotation = annotation

        self.instructions = []

    @property
    def url(self):
        return super().url + ('-async' if self.async_ else '')

    def _add_instruction(
        self,
        type_: typing.Union[str, pycamunda.instruction.ModifyInstructionType],
        activity_id: str = None,
        transition_id: str = None,
        activity_instance_id: str = None,
        transition_instance_id: str = None,
        ancestor_activity_instance_id: str = None,
        variables: typing.Mapping[str, pycamunda.variable.Variable] = None
    ) -> None:
        """Add an instruction that specify at which activities the process instance is started.

        :param type_: Type of the instruction. Possible values are
                          - startBeforeActivity
                          - startAfterActivity
                          - startTransition
                          - cancel
        :param activity_id: Id of the activity in case `type_` is `startBeforeActivity` or
                            `startAfterActivity.
        :param transition_id: Id of the sequence flow to start in case 'type_' is 'startTransition'.
        :param activity_instance_id: Id of the activity instance in case 'type_' is 'cancel'.
        :param transition_instance_id: Id of the transition instance in case 'type_' is 'cancel'.
        :param ancestor_activity_instance_id: Id of the ancestor instance id in case 'type' is
                                              'startBeforeActivity', 'startAfterActivity' or
                                              'startTransition'.
        :param variables: Mapping from names to the corresponding variables.
        """
        instruction = {'type': pycamunda.instruction.ModifyInstructionType(type_).value}
        if activity_id is not None:
            instruction['activityId'] = activity_id
        if transition_id is not None:
            instruction['transitionId'] = transition_id
        if activity_instance_id is not None:
            instruction['activityInstanceId'] = activity_instance_id
        if transition_instance_id:
            instruction['transitionInstanceId'] = transition_instance_id
        if ancestor_activity_instance_id is not None:
            instruction['ancestorActivityInstanceId'] = ancestor_activity_instance_id
        if variables is not None:
            instruction['variables'] = variables

        self.instructions.append(instruction)

    def add_before_activity_instruction(
        self,
        id_: str,
        variables: typing.Mapping[str, pycamunda.variable.Variable] = None
    ) -> None:
        """Add an instruction to start execution before a given activity is entered.

        :param id_: Id of the activity.
        :param variables: Mapping from names to the corresponding variables.
        """
        self._add_instruction(
            type_=pycamunda.instruction.ModifyInstructionType.start_before_activity,
            activity_id=id_,
            variables=variables
        )

    def add_after_activity_instruction(
        self,
        id_: str,
        variables: typing.Mapping[str, pycamunda.variable.Variable] = None
    ) -> None:
        """Add an instruction to start execution at the single outgoing sequence flow of an
        activity.

        :param id_: Id of the activity.
        :param variables: Mapping from names to the corresponding variables.
        """
        self._add_instruction(
            type_=pycamunda.instruction.ModifyInstructionType.start_after_activity,
            activity_id=id_,
            variables=variables
        )

    def add_cancel_activity_instruction(self, id_) -> None:
        """Add an instruction to cancel an activity.

        :param id_: Id of the activity.
        """
        self._add_instruction(
            type_=pycamunda.instruction.ModifyInstructionType.cancel,
            activity_id=id_
        )

    def add_transition_instruction(
        self,
        id_,
        variables: typing.Mapping[str, pycamunda.variable.Variable] = None
    ) -> None:
        """Add an instruction to start execution at the single outgoing sequence flow of an
        activity.

        :param id_: Id of the sequence flow to start.
        :param variables: Mapping from names to the corresponding variables.
        """
        self._add_instruction(
            type_=pycamunda.instruction.ModifyInstructionType.start_transition,
            transition_id=id_,
            variables=variables
        )

    def add_cancel_activity_instance_instruction(self, id_) -> None:
        """Add an instruction to cancel an activity instance.

        :param id_: Id of the activity.
        """
        self._add_instruction(
            type_=pycamunda.instruction.ModifyInstructionType.cancel,
            activity_instance_id=id_
        )

    def add_cancel_transition_instance_instruction(self, id_) -> None:
        """Add an instruction to cancel an activity instance.

        :param id_: Id of the transition instance.
        """
        self._add_instruction(
            type_=pycamunda.instruction.ModifyInstructionType.cancel,
            transition_instance_id=id_
        )

    def add_start_before_ancestor_activity_instance_instruction(
        self,
        id_: str,
        variables: typing.Mapping[str, pycamunda.variable.Variable] = None
    ) -> None:
        """Add an instruction to start execution before a given ancestor activity instance
        is entered.

        :param id_: Id of the ancestor activity instance.
        :param variables: Mapping from names to the corresponding variables.
        """
        self._add_instruction(
            type_=pycamunda.instruction.ModifyInstructionType.start_before_activity,
            ancestor_activity_instance_id=id_,
            variables=variables
        )

    def add_start_after_ancestor_activity_instance_instruction(
        self,
        id_: str,
        variables: typing.Mapping[str, pycamunda.variable.Variable] = None
    ) -> None:
        """Add an instruction to start execution at the single outgoing sequence flow of an
        ancestor activity instance.

        :param id_: Id of the ancestor activity instance.
        :param variables: Mapping from names to the corresponding variables.
        """
        self._add_instruction(
            type_=pycamunda.instruction.ModifyInstructionType.start_after_activity,
            ancestor_activity_instance_id=id_,
            variables=variables
        )

    def add_cancel_ancestor_activity_instance_instruction(self, id_: str) -> None:
        """Add an instruction to cancel an ancestor activity instance.

        :param id_: Id of the ancestor activity instance.
        """
        self._add_instruction(
            type_=pycamunda.instruction.ModifyInstructionType.cancel,
            ancestor_activity_instance_id=id_
        )

    def __call__(self, *args, **kwargs) -> typing.Optional[pycamunda.batch.Batch]:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)

        if self.async_:
            return pycamunda.batch.Batch.load(response.json())


class _ActivateSuspend(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    suspended = BodyParameter('suspended')

    def __init__(self, url: str, id_: str, suspended: bool):
        """Activate or Suspend a process definition.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the process instance.
        :param suspended: Whether to suspend or activate the process instance.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/suspended')
        self.id_ = id_
        self.suspended = suspended

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        super().__call__(pycamunda.base.RequestMethod.PUT, *args, **kwargs)


class Activate(_ActivateSuspend):

    def __init__(self, url: str, id_: str):
        """Activate a process definition.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the process definition.
        """
        super().__init__(
            url=url,
            id_=id_,
            suspended=False
        )


class Suspend(_ActivateSuspend):

    def __init__(self, url: str, id_: str):
        """Suspend a process definition.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the process definition.
        """
        super().__init__(
            url=url,
            id_=id_,
            suspended=True
        )



# -*- coding: utf-8 -*-

"""This module provides access to the deployment REST api of Camunda."""

from __future__ import annotations
import datetime as dt
import dataclasses
import typing

import requests

import pycamunda.variable
import pycamunda.base
import pycamunda.resource
import pycamunda.processdef
import pycamunda.casedef
import pycamunda.decisiondef
import pycamunda.decisionreqdef
from pycamunda.request import QueryParameter, PathParameter, BodyParameter

URL_SUFFIX = '/deployment'


__all__ = ['GetList', 'Get', 'Create', 'GetResources', 'GetResource', 'Delete']


@dataclasses.dataclass
class Deployment:
    """Data class of deployment as returned by the REST api of Camunda."""
    id_: str
    name: str
    source: str
    tenant_id: str
    deployment_time: dt.datetime = None

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> Deployment:
        deployment = cls(
            id_=data['id'],
            name=data['name'],
            source=data['source'],
            tenant_id=data['tenantId'],
        )
        if data['deploymentTime'] is not None:
            deployment.deployment_time = pycamunda.base.from_isoformat(data['deploymentTime'])

        return deployment


@dataclasses.dataclass
class Resource:
    """Data class of resource as returned by the REST api of Camunda."""
    id_: str
    name: str
    deployment_id: str

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> Resource:
        return cls(
            id_=data['id'],
            name=data['name'],
            deployment_id=data['deploymentId']
        )


@dataclasses.dataclass
class DeploymentWithDefinitions:
    """Data class of deployment with definitions as returned by the REST api of Camunda."""
    links: typing.List[pycamunda.resource.Link]
    id_: str
    name: str
    source: str
    deployed_process_definitions: typing.Dict[str, pycamunda.processdef.ProcessDefinition]
    deployed_case_definitions: typing.Dict[str, pycamunda.casedef.CaseDefinition]
    deployed_decision_definitions: typing.Dict[str, pycamunda.decisiondef.DecisionDefinition]
    deployed_decision_requirements_definitions: typing.Dict[
        str, pycamunda.decisionreqdef.DecisionRequirementsDefinition
    ]
    tenant_id: str = None
    deployment_time: dt.datetime = None

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> DeploymentWithDefinitions:
        if data['deployedProcessDefinitions'] is not None:
            deployed_process_definitions = {
                name: pycamunda.processdef.ProcessDefinition.load(process_definition)
                for name, process_definition in data['deployedProcessDefinitions'].items()
            }
        else:
            deployed_process_definitions = None
        if data['deployedCaseDefinitions'] is not None:
            deployed_case_definitions = {
                name: pycamunda.casedef.CaseDefinition.load(case_definition)
                for name, case_definition in data['deployedCaseDefinitions'].items()
            }
        else:
            deployed_case_definitions = None
        if data['deployedDecisionDefinitions'] is not None:
            deployed_decision_definitions = {
                name: pycamunda.decisiondef.DecisionDefinition.load(decision_definition)
                for name, decision_definition in data['deployedDecisionDefinitions'].items()
            }
        else:
            deployed_decision_definitions = None
        if data['deployedDecisionRequirementsDefinitions'] is not None:
            deployed_decision_requirements_definitions = {
                name: pycamunda.decisionreqdef.DecisionRequirementsDefinition.load(dec_req_def)
                for name, dec_req_def in data['deployedDecisionRequirementsDefinitions'].items()
            }
        else:
            deployed_decision_requirements_definitions = None
        return cls(
            links=[pycamunda.resource.Link.load(link) for link in data['links']],
            id_=data['id'],
            name=data['name'],
            source=data['source'],
            deployed_process_definitions=deployed_process_definitions,
            deployed_case_definitions=deployed_case_definitions,
            deployed_decision_definitions=deployed_decision_definitions,
            deployed_decision_requirements_definitions=deployed_decision_requirements_definitions,
            tenant_id=data['tenantId'],
            deployment_time=pycamunda.base.from_isoformat(data['deploymentTime'])
        )


class GetList(pycamunda.base.CamundaRequest):

    id_ = QueryParameter('id')
    name = QueryParameter('name')
    name_like = QueryParameter('nameLike')
    source = QueryParameter('source')
    without_source = QueryParameter('withoutSource')
    tenant_id_in = QueryParameter('tenantIdIn')
    without_tenant_id = QueryParameter('withoutTenantId', provide=pycamunda.base.value_is_true)
    include_deployments_without_tenant_id = QueryParameter(
        'includeDeploymentsWithoutTenantId', provide=pycamunda.base.value_is_true
    )
    after = QueryParameter('after')
    before = QueryParameter('before')
    sort_by = QueryParameter(
        'sortBy',
        mapping={
            'id_': 'id',
            'name': 'name',
            'definition_time': 'definitionTime',
            'tenant_id': 'tenantId',
        }
    )
    ascending = QueryParameter(
        'sortOrder',
        mapping={True: 'asc', False: 'desc'},
        provide=lambda self, obj, obj_type: vars(obj).get('sort_by', None) is not None
    )
    first_result = QueryParameter('firstResult')
    max_results = QueryParameter('maxResults')

    def __init__(
        self,
        url: str,
        id_: str = None,
        name: str = None,
        name_like: str = None,
        source: str = None,
        without_source: bool = None,
        tenant_id_in: typing.Iterable[str] = None,
        without_tenant_id: bool = False,
        include_deployments_without_tenant_id: bool = False,
        after: dt.datetime = None,
        before: dt.datetime = None,
        sort_by: str = None,
        ascending: bool = True,
        first_result: int = None,
        max_results: int = None,
    ):
        """Get a list of deployments.
        
        :param url: Camunda Rest engine URL.
        :param id_: Filter by id.
        :param name: Filter by name.
        :param name_like: Filter by a substring of the name.
        :param source: Filter by deployment source.
        :param without_source: Filter deployments without source.
        :param tenant_id_in: Filter whether the tenant id is one of multiple ones.
        :param without_tenant_id: Whether to include only deployments that belong to no tenant.
        :param include_deployments_without_tenant_id: Whether to include deployments that belong to
                                                      no tenant.
        :param after: Filter by the deployment date after the provided data.
        :param before: Filter by the deployment date before the provided data.
        :param sort_by: Sort the results by 'id', 'name', 'deployment_time' or 'tenant_id'.
        :param ascending: Sort order.
        :param first_result: Pagination of results. Index of the first result to return.
        :param max_results: Pagination of results. Maximum number of results to return.
        """
        super().__init__(url=url + URL_SUFFIX)
        self.id_ = id_
        self.name = name
        self.name_like = name_like
        self.source = source
        self.without_source = without_source
        self.tenant_id_in = tenant_id_in
        self.without_tenant_id = without_tenant_id
        self.include_deployments_without_tenant_id = include_deployments_without_tenant_id
        self.after = after
        self.before = before
        self.sort_by = sort_by
        self.ascending = ascending
        self.first_result = first_result
        self.max_results = max_results

    def __call__(self, *args, **kwargs) -> typing.Tuple[Deployment]:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return tuple(Deployment.load(deployment_json) for deployment_json in response.json())


class Get(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')

    def __init__(self, url: str, id_: str):
        """Get a deployment.

        :param url: Camunda Rest engine url.
        :param id_: Id of the deployment.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}')
        self.id_ = id_

    def __call__(self, *args, **kwargs) -> Deployment:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return Deployment.load(response.json())


class Create(pycamunda.base.CamundaRequest):

    name = BodyParameter('deployment-name')
    enable_duplicate_filtering = BodyParameter('enable-duplicate-filtering')
    deploy_changed_only = BodyParameter('deploy-changed-only')
    source = BodyParameter('deployment-source')
    tenant_id = BodyParameter('tenant-id')

    def __init__(
        self,
        url: str,
        name: str,
        source: str = None,
        enable_duplicate_filtering: bool = False,
        deploy_changed_only: bool = False,
        tenant_id: str = None
    ):
        """Create a deployment with one or multiple resources (e.g. bpmn diagrams).

        :param url: Camunda Rest engine url
        :param name: Name of the deployment.
        :param source: Source of the deployment.
        :param enable_duplicate_filtering: Whether to check if a deployment with the same name
                                           already exists. If one does, no new deployment is done.
        :param deploy_changed_only: Whether to check if already deployed resources that are also
                                    contained in this deployment have changed. The ones that do not
                                    will not created again.
        :param tenant_id: Id of the tenant to create the deployment for.
        """
        super().__init__(url=url + URL_SUFFIX + '/create')
        self.name = name
        self.source = source
        self.enable_duplicate_filtering = enable_duplicate_filtering
        self.deploy_changed_only = deploy_changed_only
        self.tenant_id = tenant_id

        self._files = []

    def add_resource(self, file) -> None:
        """Add a resource to the deployment.

        :param file: Binary data of the resource.
        """
        self._files.append(file)

    @property
    def files(self):
        return {f'resource-{i}': resource for i, resource in enumerate(self._files)}

    def __call__(self, *args, **kwargs) -> DeploymentWithDefinitions:
        """Send the request."""
        assert bool(self.files), 'Cannot create deployment without resources.'
        try:
            response = requests.request(
                method=pycamunda.base.RequestMethod.POST.value,
                url=self.url,
                params=self.query_parameters(),
                data=self.body_parameters(),
                auth=self.auth,
                files=self.files
            )
        except requests.exceptions.RequestException as exc:
            raise pycamunda.PyCamundaException(exc)
        if not response:
            pycamunda.base._raise_for_status(response)

        return DeploymentWithDefinitions.load(data=response.json())


class GetResources(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')

    def __init__(self, url: str, id_: str):
        """Get the resources of a deployment.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the the deployment.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/resources')
        self.id_ = id_

    def __call__(self, *args, **kwargs) -> typing.Tuple[Resource]:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return tuple(Resource.load(resource_json) for resource_json in response.json())


class GetResource(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    resource_id = PathParameter('resourceId')

    def __init__(self, url: str, id_: str, resource_id: str, binary: bool = False):
        """Get a resource of a deployment.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the the deployment.
        :param resource_id: Id of the resource.
        :param binary: Whether to request binary content of the resource.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/resources/{resourceId}')
        self.id_ = id_
        self.resource_id = resource_id
        self.binary = binary

    @property
    def url(self):
        return super().url + ('/data' if self.binary else '')

    def __call__(self, *args, **kwargs) -> typing.Union[Resource, typing.ByteString]:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        if self.binary:
            return response.content
        return Resource.load(response.json())


class Delete(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    cascade = QueryParameter('cascade')
    skip_custom_listeners = QueryParameter('skipCustomListeners')
    skip_io_mappings = QueryParameter('skipIoMappings')

    def __init__(
        self,
        url: str,
        id_: str,
        cascade: bool = False,
        skip_custom_listeners: bool = False,
        skip_io_mappings: bool = False
    ):
        """Delete a deployment.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the the deployment.
        :param cascade: Whether to cascade the action to all process instances (including historic).
        :param skip_custom_listeners: Whether to skip custom listeners and notify only builtin ones.
        :param skip_io_mappings: Whether to skip input/output mappings.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}')
        self.id_ = id_
        self.cascade = cascade
        self.skip_custom_listeners = skip_custom_listeners
        self.skip_io_mappings = skip_io_mappings

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        super().__call__(pycamunda.base.RequestMethod.DELETE, *args, **kwargs)



# -*- coding: utf-8 -*-

"""This module provides access to the group REST api of Camunda."""

from __future__ import annotations
import dataclasses
import typing

import pycamunda
import pycamunda.resource
import pycamunda.variable
import pycamunda.base
from pycamunda.request import QueryParameter, PathParameter, BodyParameter

URL_SUFFIX = '/group'
URL_SUFFIX_MEMBERS = '/members'


__all__ = [
    'Get', 'GetList', 'Create', 'Update', 'Options', 'Delete', 'MemberCreate', 'MemberDelete',
    'MemberOptions'
]


@dataclasses.dataclass
class Group:
    """Data class of group as returned by the REST api of Camunda."""
    id_: str
    name: str
    type_: str = None

    @classmethod
    def load(cls, data) -> Group:
        group = cls(
            id_=data['id'],
            name=data['name']
        )
        try:
            group.type_ = data['type']
        except KeyError:
            pass

        return group


class Get(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')

    def __init__(self, url: str, id_: str):
        """Get a group.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the group.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}')
        self.id_ = id_

    def __call__(self, *args, **kwargs) -> Group:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return Group.load(response.json())


class GetList(pycamunda.base.CamundaRequest):

    id_ = QueryParameter('id')
    id_in = QueryParameter('idIn')
    name = QueryParameter('name')
    name_like = QueryParameter('nameLike')
    type_ = QueryParameter('type')
    member = QueryParameter('member')
    member_of_tenant = QueryParameter('memberOfTenant')
    sort_by = QueryParameter('sortBy', mapping={'id_': 'id', 'name': 'name', 'type_': 'type'})
    ascending = QueryParameter(
        'sortOrder',
        mapping={True: 'asc', False: 'desc'},
        provide=lambda self, obj, obj_type: vars(obj).get('sort_by', None) is not None
    )
    first_result = QueryParameter('firstResult')
    max_results = QueryParameter('maxResults')

    def __init__(
        self,
        url,
        id_: str = None,
        id_in: typing.Iterable[str] = None,
        name: str = None,
        name_like: str = None,
        type_: str = None,
        member: str = None,
        member_of_tenant: str = None,
        sort_by: str = None,
        ascending: bool = True,
        first_result: int = None,
        max_results: int = None
    ):
        """Get a list of groups.

        :param url: Camunda Rest engine URL.
        :param id_: Filter by the id of the group.
        :param id_in: Filter whether the id of the group is one of multiple ones.
        :param name: Filter by the name of the group.
        :param name_like: Filter by a substring of the name of the group.
        :param type_: Filter by the type of the group.
        :param member: Filter by a member of the group.
        :param member_of_tenant: Filter by member of the tenant.
        :param sort_by: Sort the results by `id_`, `name` or `type_`.
        :param ascending: Sort order.
        :param first_result: Pagination of results. Index of the first result to return.
        :param max_results: Pagination of results. Maximum number of results to return.
        """
        super().__init__(url=url + URL_SUFFIX)
        self.id_ = id_
        self.id_in = id_in
        self.name = name
        self.name_like = name_like
        self.type_ = type_
        self.member = member
        self.member_of_tenant = member_of_tenant
        self.sort_by = sort_by
        self.ascending = ascending
        self.first_result = first_result
        self.max_results = max_results

    def __call__(self, *args, **kwargs) -> typing.Tuple[Group]:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return tuple(Group.load(group_json) for group_json in response.json())


class Create(pycamunda.base.CamundaRequest):

    id_ = BodyParameter('id')
    name = BodyParameter('name')
    type_ = BodyParameter('type')

    def __init__(self, url: str, id_: str, name: str, type_: str):
        """Create a new group.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the group.
        :param name: Name of the group.
        :param type_: Type of the group.
        """
        super().__init__(url=url + URL_SUFFIX + '/create')
        self.id_ = id_
        self.name = name
        self.type_ = type_

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)


class Update(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    name = BodyParameter('name')
    type_ = BodyParameter('type')

    def __init__(self, url: str, id_: str, name: str, type_: str):
        """Update a group.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the group.
        :param name: New name of the group.
        :param type_: New zype of the group.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}')
        self.id_ = id_
        self.name = name
        self.type_ = type_

    def body_parameters(self, apply: typing.Callable = ...):
        params = super().body_parameters(apply=apply)
        params['id'] = self.id_
        return params

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        super().__call__(pycamunda.base.RequestMethod.PUT, *args, **kwargs)


class Options(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')

    def __init__(self, url: str, id_=None):
        """Get a list of options the currently authenticated user can perform on the group resource.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the user.
        """
        super().__init__(url=url + URL_SUFFIX)
        self.id_ = id_

    @property
    def url(self):
        return super().url + (f'/{self.id_}' if self.id_ is not None else '')

    def __call__(self, *args, **kwargs):
        """Send the request"""
        response = super().__call__(pycamunda.base.RequestMethod.OPTIONS, *args, **kwargs)

        return pycamunda.resource.ResourceOptions.load(response.json())


class Delete(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')

    def __init__(self, url: str, id_: str):
        """Delete a group.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the group.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}')
        self.id_ = id_

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.DELETE, *args, **kwargs)


class MemberCreate(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    user_id = PathParameter('userId')

    def __init__(self, url: str, id_: str, user_id: str):
        """Add a member to a group.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the group.
        :param user_id: Id of the user.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}' + URL_SUFFIX_MEMBERS + '/{userId}')
        self.id_ = id_
        self.user_id = user_id

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        super().__call__(pycamunda.base.RequestMethod.PUT, *args, **kwargs)


class MemberDelete(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    user_id = PathParameter('userId')

    def __init__(self, url: str, id_: str, user_id: str):
        """Delete a member from a group.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the group.
        :param user_id: Id of the user.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}' + URL_SUFFIX_MEMBERS + '/{userId}')
        self.id_ = id_
        self.user_id = user_id

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        super().__call__(pycamunda.base.RequestMethod.DELETE, *args, **kwargs)


class MemberOptions(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')

    def __init__(self, url: str, id_: str):
        """Get a list of options the currently authenticated user can perform on the group member
        resource.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the group.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}' + URL_SUFFIX_MEMBERS)
        self.id_ = id_

    def __call__(self, *args, **kwargs) -> pycamunda.resource.ResourceOptions:
        """Send the request."""
        response = super().__call__(pycamunda.base.RequestMethod.OPTIONS, *args, **kwargs)

        return pycamunda.resource.ResourceOptions.load(response.json())



# -*- coding: utf-8 -*-

"""This module provides access to the batch REST api of Camunda"""

from __future__ import annotations
import dataclasses
import typing


__all__ = []


@dataclasses.dataclass
class Batch:
    """Data class of batch as returned by the REST api of Camunda."""
    id_: str
    type_: str
    total_jobs: int
    jobs_created: int
    batch_jobs_per_seed: int
    invocations_per_batch_job: int
    seed_job_definition_id: str
    monitor_job_definition_id: str
    batch_job_definition_id: str
    suspended: bool
    tenant_id: str
    create_user_id: str

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> Batch:
        return cls(
            id_=data['id'],
            type_=data['type'],
            total_jobs=data['totalJobs'],
            jobs_created=data['jobsCreated'],
            batch_jobs_per_seed=data['batchJobsPerSeed'],
            invocations_per_batch_job=data['invocationsPerBatchJob'],
            seed_job_definition_id=data['seedJobDefinitionId'],
            monitor_job_definition_id=data['monitorJobDefinitionId'],
            batch_job_definition_id=data['batchJobDefinitionId'],
            suspended=data['suspended'],
            tenant_id=data['tenantId'],
            create_user_id=data['createUserId']
        )



# -*- coding: utf-8 -*-

"""This module provides access to the user REST api of Camunda."""

from __future__ import annotations
import dataclasses
import typing

import pycamunda
import pycamunda.resource
import pycamunda.variable
import pycamunda.base
from pycamunda.request import QueryParameter, PathParameter, BodyParameter, BodyParameterContainer

URL_SUFFIX = '/user'


__all__ = [
    'Delete', 'Count', 'GetList', 'GetProfile', 'Options', 'Create', 'UpdateCredentials',
    'UpdateProfile', 'Unlock'
]


@dataclasses.dataclass
class User:
    """Data class of user as returned by the REST api of Camunda."""
    id_: str
    first_name: str
    last_name: str
    email: str = None
    display_name = None

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> User:
        user = cls(
            id_=data['id'],
            first_name=data['firstName'],
            last_name=data['lastName']
        )
        try:
            user.email = data['email']
        except KeyError:
            pass
        try:
            user.display_name = data['displayName']
        except KeyError:
            pass

        return user


class Delete(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')

    def __init__(self, url: str, id_: str):
        """Deletes a user by id.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the user.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}')
        self.id_ = id_

    def __call__(self, *args, **kwargs) -> None:
        """Send the request"""
        super().__call__(pycamunda.base.RequestMethod.DELETE, *args, **kwargs)


class Count(pycamunda.base.CamundaRequest):

    id_ = QueryParameter('id')
    first_name = QueryParameter('firstName')
    first_name_like = QueryParameter('firstNameLike')
    last_name = QueryParameter('lastName')
    last_name_like = QueryParameter('lastNameLike')
    email = QueryParameter('email')
    email_like = QueryParameter('emailLike')
    member_of_group = QueryParameter('memberOfGroup')
    member_of_tenant = QueryParameter('memberOfTenant')

    def __init__(
        self,
        url: str,
        id_: str = None,
        first_name: str = None,
        first_name_like: str = None,
        last_name: str = None,
        last_name_like: str = None,
        email: str = None,
        email_like: str = None,
        member_of_group: str = None,
        member_of_tenant: str = None
    ):
        """Count users.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the user.
        :param first_name: Filter by the first name of the user.
        :param first_name_like: Filter by a substring of the first name.
        :param last_name: Filter by the last name of the user.
        :param last_name_like: Filter by a substring of the last name.
        :param email: Filter by the email of the user.
        :param email_like: Filter by a substring of the email.
        :param member_of_group: Filter for users which are a member of a group.
        :param member_of_tenant: Filter for users which are a member of a tenant.
        """
        super().__init__(url=url + URL_SUFFIX + '/count')
        self.id_ = id_
        self.first_name = first_name
        self.first_name_like = first_name_like
        self.last_name = last_name
        self.last_name_like = last_name_like
        self.email = email
        self.email_like = email_like
        self.member_of_group = member_of_group
        self.member_of_tenant = member_of_tenant

    def __call__(self, *args, **kwargs) -> int:
        """Send the request"""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return response.json()['count']


class GetList(pycamunda.base.CamundaRequest):

    id_ = QueryParameter('id')
    first_name = QueryParameter('firstName')
    first_name_like = QueryParameter('firstNameLike')
    last_name = QueryParameter('lastName')
    last_name_like = QueryParameter('lastNameLike')
    email = QueryParameter('email')
    email_like = QueryParameter('emailLike')
    member_of_group = QueryParameter('memberOfGroup')
    member_of_tenant = QueryParameter('memberOfTenant')
    sort_by = QueryParameter(
        'sortBy',
        mapping={
            'id_': 'userId',
            'first_name': 'firstName',
            'last_name': 'lastName',
            'email': 'email'
        }
    )
    ascending = QueryParameter(
        'sortOrder',
        mapping={True: 'asc', False: 'desc'},
        provide=lambda self, obj, obj_type: vars(obj).get('sort_by', None) is not None
    )
    first_result = QueryParameter('firstResult')
    max_results = QueryParameter('maxResults')

    def __init__(
        self,
        url: str,
        id_: str = None,
        first_name: str = None,
        first_name_like: str = None,
        last_name: str = None,
        last_name_like: str = None,
        email: str = None,
        email_like: str = None,
        member_of_group: str = None,
        member_of_tenant: str = None,
        sort_by: str = None,
        ascending: bool = True,
        first_result: int = None,
        max_results: int = None
    ):
        """Query for a list of users using a list of parameters. The size of the result set can be
        retrieved by using the Get User Count method.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the user.
        :param first_name: Filter by the first name of the user.
        :param first_name_like: Filter by a substring of the first name.
        :param last_name: Filter by the last name of the user.
        :param last_name_like: Filter by a substring of the last name.
        :param email: Filter by the email of the user.
        :param email_like: Filter by a substring of the email.
        :param member_of_group: Filter for users which are a member of a group.
        :param member_of_tenant: Filter for users which are a member of a tenant.
        :param sort_by: Sort the results by `id_`, `first_name`, `last_name` or `email` of the user.
        :param ascending: Sort order.
        :param first_result: Pagination of results. Index of the first result to return.
        :param max_results: Pagination of results. Maximum number of results to return.
        """
        super().__init__(url=url + URL_SUFFIX)
        self.id_ = id_
        self.first_name = first_name
        self.first_name_like = first_name_like
        self.last_name = last_name
        self.last_name_like = last_name_like
        self.email = email
        self.email_like = email_like
        self.member_of_group = member_of_group
        self.member_of_tenant = member_of_tenant
        self.sort_by = sort_by
        self.ascending = ascending
        self.first_result = first_result
        self.max_results = max_results

    def __call__(self, *args, **kwargs) -> typing.Tuple[User]:
        """Send the request"""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return tuple(User.load(user_json) for user_json in response.json())


class GetProfile(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')

    def __init__(self, url: str, id_: str):
        """Get the profile of an user.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the user.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/profile')
        self.id_ = id_

    def __call__(self, *args, **kwargs) -> User:
        """Send the request"""
        response = super().__call__(pycamunda.base.RequestMethod.GET, *args, **kwargs)

        return User.load(response.json())


class Options(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')

    def __init__(self, url: str, id_: str = None):
        """Get a list of options the currently authenticated user can perform on the user resource.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the user.
        """
        super().__init__(url=url + URL_SUFFIX)
        self.id_ = id_

    @property
    def url(self):
        return super().url + (f'/{self.id_}' if self.id_ is not None else '')

    def __call__(self, *args, **kwargs) -> pycamunda.resource.ResourceOptions:
        """Send the request"""
        response = super().__call__(pycamunda.base.RequestMethod.OPTIONS, *args, **kwargs)

        return pycamunda.resource.ResourceOptions.load(response.json())


class Create(pycamunda.base.CamundaRequest):

    id_ = BodyParameter('id')
    first_name = BodyParameter('firstName')
    last_name = BodyParameter('lastName')
    email = BodyParameter('email')
    profile = BodyParameterContainer('profile', id_, first_name, last_name, email)

    password = BodyParameter('password')
    credentials = BodyParameterContainer('credentials', password)

    def __init__(
        self,
        url: str,
        id_: str = None,
        first_name: str = None,
        last_name: str = None,
        email: str = None,
        password: str = None
    ):
        """Create a new user.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the user.
        :param first_name: First name of the user.
        :param last_name: Last name of the user.
        :param email: Email of the user.
        :param password: Password of the user.
        """
        super().__init__(url=url + URL_SUFFIX + '/create')
        self.id_ = id_
        self.first_name = first_name
        self.last_name = last_name
        self.email = email
        self.password = password

    def __call__(self, *args, **kwargs) -> None:
        """Send the request"""
        super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)


class UpdateCredentials(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    password = BodyParameter('password')
    authenticated_user_password = BodyParameter('authenticatedUserPassword')

    def __init__(self, url: str, id_: str, password: str, authenticated_user_password: str):
        """Update a user's credentials (password).

        :param url: Camunda Rest engine URL.
        :param id_: Id of the user.
        :param password: New password of the user.
        :param authenticated_user_password: Password of the authenticated user who changes the
                                            password.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/credentials')
        self.id_ = id_
        self.password = password
        self.authenticated_user_password = authenticated_user_password

    def __call__(self, *args, **kwargs) -> None:
        """Send the request"""
        super().__call__(pycamunda.base.RequestMethod.PUT, *args, **kwargs)


class UpdateProfile(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')
    new_user_id = BodyParameter('id')
    first_name = BodyParameter('firstName')
    last_name = BodyParameter('lastName')
    email = BodyParameter('email')

    def __init__(
        self,
        url: str,
        id_: str,
        new_user_id: str = None,
        first_name: str = None,
        last_name: str = None,
        email: str = None
    ):
        """Update the profile information of an already existing user.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the user.
        :param new_user_id: New user id of the user.
        :param first_name: New first name of the user.
        :param last_name: New last name of the user.
        :param email: New email of the user.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/profile')
        self.id_ = id_
        self.new_user_id = new_user_id
        self.first_name = first_name
        self.last_name = last_name
        self.email = email

    def __call__(self, *args, **kwargs) -> None:
        """Send the request"""
        super().__call__(pycamunda.base.RequestMethod.PUT, *args, **kwargs)


class Unlock(pycamunda.base.CamundaRequest):

    id_ = PathParameter('id')

    def __init__(
        self,
        url: str,
        id_: str = None
    ):
        """Unlock an user.

        :param url: Camunda Rest engine URL.
        :param id_: Id of the user.
        """
        super().__init__(url=url + URL_SUFFIX + '/{id}/unlock')
        self.id_ = id_

    def __call__(self, *args, **kwargs) -> None:
        """Send the request"""
        super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)



# -*- coding: utf-8 -*-

import abc
import typing


__all__ = []


class InvalidInput(ValueError):
    """Exception that is raised when invalid input is provided."""


class RequestParameter:

    def __init__(
        self,
        key: str,
        mapping: typing.Mapping = None,
        provide: typing.Callable = None,
        validate: typing.Callable = None
    ):
        """Parameter that is send with a Request when it is attached to the class and its value is
        set. This class implements the descriptor protocol.

        :param key: Api key of the request parameter.
        :param mapping: Mapping from descriptor value to the parameter that is send.
        :param provide: Callable that determines whether the value is returned. Is expected to
                        accept 3 arguments:
                            - Descriptor instance,
                            - Object instance the descriptor is attached to and
                            - Type of the object the descriptor is attached to.
        :param validate: Callable that validates the value that is tried to be set.
        """
        self.key = key
        self.mapping = mapping
        self.provide = provide
        self.validate = validate
        self.name = None

    def __get__(self, obj: typing.Any, obj_type: typing.Any = None) -> typing.Any:
        if self.provide is None or self.provide(self, obj, obj_type):
            if self.mapping is None:
                return obj.__dict__[self.name]
            return self.mapping[obj.__dict__[self.name]]

    def __set__(self, obj: typing.Any, value: typing.Any):
        if self.validate is not None and not self.validate(value):
            raise InvalidInput(f'Cannot set value "{value}" for "{self.name}"')
        obj.__dict__[self.name] = value

    def __repr__(self) -> str:
        return f'{self.__class__.__qualname__}(key=\'{self.key}\')'


class QueryParameter(RequestParameter):
    """Parameter that is attached to the request URL by adding it after the endpoint name."""


class PathParameter(RequestParameter):
    """Parameter that is attached to the request URL by adding it to the endpoint name."""


class BodyParameter(RequestParameter):

    def __init__(self, *args, **kwargs):
        """Parameter that is attached to the request body."""
        super().__init__(*args, **kwargs)
        self.hidden = False


class BodyParameterContainer:
    """Stores multiple BodyParameters and allows sending nested queries.

    :param key: Key for the nested query.
    :param args: BodyParameter`s
    """
    def __init__(self, key: str, *parameters):
        self.key = key
        self.parameters = {}
        for parameter in parameters:
            self.parameters[parameter.key] = parameter
            parameter.hidden = True

    def __repr__(self) -> str:
        return f'{self.__class__.__qualname__}(key={self.key}, ' \
               f'{", ".join(k+"="+str(v) for k, v in self.parameters.items())})'


class RequestMeta(abc.ABCMeta):

    def __init__(cls, name: str, bases: typing.Any, attr_dict: typing.Dict[str, typing.Any]):

        super().__init__(name, bases, attr_dict)

        try:
            cls._parameters = dict(cls._parameters)
        except AttributeError:
            cls._parameters = {}
        try:
            cls._containers = dict(cls._containers)
        except AttributeError:
            cls._containers = {}

        for key, attr in attr_dict.items():
            if isinstance(attr, RequestParameter):
                attr.name = key
                cls._parameters[key] = attr
            elif isinstance(attr, BodyParameterContainer):
                cls._containers[key] = attr


class Request(metaclass=RequestMeta):

    def __init__(self, url: str):
        """Abstract base class for requests. Extracts parameters to send with the requests by
        parsing the class for RequestParameter`s.

        :param url: Rest engine url.
        """
        super().__init__()
        self._url = url

    @property
    def url(self) -> str:
        params = {}
        missing_params = {}
        for name, attribute in self._parameters.items():
            if isinstance(attribute, PathParameter):
                try:
                    params[attribute.key] = getattr(self, attribute.name)
                except AttributeError:
                    missing_params[attribute.key] = ''
        return self._url.format(**{**params, **missing_params}).rstrip('/')

    @abc.abstractmethod
    def __call__(self, *args, **kwargs):
        return NotImplementedError

    def query_parameters(self, apply: typing.Callable = None) -> typing.Dict[str, typing.Any]:
        query = {}
        for name, attribute in self._parameters.items():
            if isinstance(attribute, QueryParameter):
                try:
                    value = getattr(self, attribute.name)
                except KeyError:
                    pass
                else:
                    if value is not None:
                        query[attribute.key] = value

        if apply is None:
            return query
        return {key: apply(value) for key, value in query.items()}

    def _traverse(self, container: BodyParameterContainer) -> typing.Dict[str, typing.Any]:
        query = {}
        for key, val in container.parameters.items():
            if isinstance(val, BodyParameterContainer):
                query[key] = self._traverse(val)
            else:
                try:
                    value = getattr(self, val.name)
                except KeyError:
                    pass
                except AttributeError:
                    if val is not None:
                        query[key] = val
                else:
                    if value is not None:
                        query[key] = value
        return query

    def body_parameters(self, apply: typing.Callable = None) -> typing.Dict[str, typing.Any]:
        query = {}
        for name, attribute in self._containers.items():
            if isinstance(attribute, BodyParameterContainer):
                query[attribute.key] = self._traverse(attribute)
        for name, attribute in self._parameters.items():
            if isinstance(attribute, BodyParameter) and not attribute.hidden:
                try:
                    value = getattr(self, attribute.name)
                except KeyError:
                    pass
                else:
                    if value is not None:
                        query[attribute.key] = value

        if apply is None:
            return query
        return {key: apply(value) for key, value in query.items()}

    def __repr__(self) -> str:
        keys_values = {}
        for param in self.__class__._parameters.keys():
            try:
                value = getattr(self, param)
            except KeyError:
                pass
            else:
                if value is not None:
                    keys_values[param] = value
        args = [f"url={self.url!r}"] + [f"{key}={value!r}" for key, value in keys_values.items()]
        return f'{self.__class__.__qualname__}({", ".join(args)})'



# -*- coding: utf-8 -*-

"""This module provides access to the decision requirements definition REST api of Camunda."""

from __future__ import annotations
import typing
import dataclasses


@dataclasses.dataclass
class DecisionRequirementsDefinition:
    """Data class of decision requirements definition as returned by the REST api of Camunda."""
    id_: str
    key: str
    category: str
    name: str
    version: int
    resource: str
    deployment_id: str
    tenant_id: str

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> DecisionRequirementsDefinition:
        return cls(
            id_=data['id'],
            key=data['key'],
            category=data['category'],
            name=data['name'],
            version=data['version'],
            resource=data['resource'],
            deployment_id=data['deploymentId'],
            tenant_id=data['tenantId']
        )



# -*- coding: utf-8 -*-

"""This module provides instruction enumerations as returned by the REST api of Camunda."""

import enum


__all__ = ['InstructionType', 'ModifyInstructionType']


class InstructionType(enum.Enum):
    start_before_activity = 'startBeforeActivity'
    start_after_activity = 'startAfterActivity'
    start_transition = 'startTransition'


class ModifyInstructionType(enum.Enum):
    start_before_activity = 'startBeforeActivity'
    start_after_activity = 'startAfterActivity'
    start_transition = 'startTransition'
    cancel = 'cancel'



# -*- coding: utf-8 -*-

"""This module provides activity instance data classes as returned by the REST api of Camunda."""

from __future__ import annotations
import dataclasses
import typing

import pycamunda.incident


__all__ = []


@dataclasses.dataclass
class TransitionInstance:
    """Data class of transition instance as returned by the REST api of Camunda."""
    id_: str
    activity_id: str
    activity_name: str
    activity_type: str
    process_instance_id: str
    process_definition_id: str
    execution_ids: typing.Tuple[str]
    incident_ids: typing.Tuple[str]
    incidents: typing.Tuple[pycamunda.incident.IncidentTypeCount]

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> TransitionInstance:
        return cls(
            id_=data['id'],
            activity_id=data['activityId'],
            activity_name=data['activityName'],
            activity_type=data['activityType'],
            process_instance_id=data['processInstanceId'],
            process_definition_id=data['processDefinitionId'],
            execution_ids=tuple(data['executionId']),
            incident_ids=tuple(data['incidentIds']),
            incidents=tuple(
                pycamunda.incident.IncidentTypeCount.load(incident_json)
                for incident_json in data['incidents'])
        )


@dataclasses.dataclass
class ActivityInstance:
    """Data class of activity instance as returned by the REST api of Camunda."""
    id_: str
    parent_activity_instance_id: str
    activity_id: str
    activity_name: str
    activity_type: str
    process_instance_id: str
    process_definition_id: str
    child_activity_instances: typing.Tuple[ActivityInstance]
    child_transition_instances: typing.Tuple[TransitionInstance]
    execution_ids: typing.Tuple[str]
    name: str
    incident_ids: typing.Tuple[str] = None
    incidents: typing.Tuple[pycamunda.incident.IncidentTypeCount] = None

    @classmethod
    def load(cls, data: typing.Mapping[str, typing.Any]) -> ActivityInstance:
        activity_instance = cls(
            id_=data['id'],
            parent_activity_instance_id=data['parentActivityInstanceId'],
            activity_id=data['activityId'],
            activity_name=data['activityName'],
            activity_type=data['activityType'],
            process_instance_id=data['processInstanceId'],
            process_definition_id=data['processDefinitionId'],
            child_activity_instances=tuple(
                cls.load(activity_json) for activity_json in data['childActivityInstances']
            ),
            child_transition_instances=tuple(
                TransitionInstance.load(transition_json)
                for transition_json in data['childTransitionInstances']
            ),
            execution_ids=tuple(data['executionIds']),
            name=data['name']
        )
        try:
            activity_instance.incident_ids = tuple(data['incidentIds'])
        except KeyError:
            pass
        try:
            activity_instance.incidents = tuple(
                pycamunda.incident.IncidentTypeCount.load(incident_json)
                for incident_json in data['incidents']
            )
        except KeyError:
            pass

        return activity_instance



# -*- coding: utf-8 -*-

import enum
import datetime as dt
import typing
import json

import requests

import pycamunda.request


__all__ = ['isoformat', 'from_isoformat']


def value_is_true(self, obj: typing.Any, obj_type: typing.Any) -> bool:
    return bool(obj.__dict__[self.name])


class RequestMethod(enum.Enum):
    GET = 'GET'
    POST = 'POST'
    PUT = 'PUT'
    PATCH = 'PATCH'
    DELETE = 'DELETE'
    OPTIONS = 'OPTIONS'
    HEAD = 'HEAD'


class CamundaRequest(pycamunda.request.Request):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.auth = None
        self._files = None

    @property
    def files(self):
        return self._files

    def __call__(self, method: RequestMethod, *args, **kwargs) -> requests.Response:
        try:
            response = requests.request(
                method=method.value,
                url=self.url,
                params=self.query_parameters(),
                json=self.body_parameters(),
                auth=self.auth,
                files=self.files
            )
        except requests.exceptions.RequestException as exc:
            raise pycamunda.PyCamundaException(exc)
        if not response:
            pycamunda.base._raise_for_status(response)

        return response

    def body_parameters(self, apply: typing.Callable = ...):
        if apply is Ellipsis:
            return super().body_parameters(apply=prepare)
        return super().body_parameters(apply=apply)

    def query_parameters(self, apply: typing.Callable = ...):
        if apply is Ellipsis:
            return super().query_parameters(apply=query_prepare)
        return super().query_parameters(apply=apply)


def isoformat(datetime_: typing.Union[dt.date, dt.datetime]) -> str:
    """Convert a datetime object to the isoformat string Camunda expects. Datetime objects are
    expected to contain timezoneinformation.

    :param datetime_: Datetime or date object to convert.
    :return: Isoformat datetime or date string.
    """
    if isinstance(datetime_, dt.datetime):
        dt_str = datetime_.strftime('%Y-%m-%dT%H:%M:%S.{ms}%z')
        ms = datetime_.microsecond // 1000
        dt_str = dt_str.format(ms=str(ms).zfill(3))
    else:
        dt_str = datetime_.strftime('%Y-%m-%d')

    return dt_str


def from_isoformat(datetime_str: str) -> dt.datetime:
    """Convert an isoformat string to a datetime object.

    :param datetime_str: String to convert.
    :return: Converted datetime.
    """
    return dt.datetime.strptime(datetime_str, '%Y-%m-%dT%H:%M:%S.%f%z')


def prepare(value: typing.Any) -> typing.Any:
    """Prepare parameter values for Camunda.

    :param value: Value to prepare.
    """
    if isinstance(value, dt.datetime):
        return isoformat(datetime_=value)
    try:  # support for enums
        value = value.value
    except AttributeError:
        pass
    return value


def query_prepare(value: typing.Any) -> typing.Any:
    """Prepare parameter values for Camunda including converting boolean values to lowercase
    strings.

    :param value: Value to prepare.
    """
    if isinstance(value, bool):
        return 'true' if value else 'false'
    return prepare(value)


def _raise_for_status(response: requests.Response) -> None:
    """Handle http errors regarding Camunda.

    :param response: Response coming from Camunda.
    """
    try:
        message = response.json()['message']
    except KeyError:
        message = ''
    except TypeError:
        message = ''
    except json.decoder.JSONDecodeError:
        message = ''

    if response.status_code == 400:
        raise pycamunda.BadRequest(message)
    if response.status_code == 401:
        raise pycamunda.Unauthorized(message)
    elif response.status_code == 403:
        raise pycamunda.Forbidden(message)
    elif response.status_code == 404:
        raise pycamunda.NotFound(message)

    raise pycamunda.NoSuccess(response.text)



# !/usr/bin/env python
# -*- coding: utf-8 -*-

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='pycamunda',
    version='0.1.0',
    author='Peter Klauke',
    description='A high-level framework for communicating with the workflow and decision automation engine Camunda.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/pklauke/pycamunda',
    packages=['pycamunda'],
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=['requests>=2.0.0']
)



# -*- coding: utf-8 -*-

"""This module provides access to the signal REST api of Camunda."""

from __future__ import annotations
import typing

import pycamunda.variable
import pycamunda.base
from pycamunda.request import BodyParameter

URL_SUFFIX = '/signal'


__all__ = ['EventAll', 'EventSingle']


class _Event(pycamunda.base.CamundaRequest):

    name = BodyParameter('name')
    execution_id = BodyParameter('executionId')
    variables = BodyParameter('variables')
    tenant_id = BodyParameter('tenantId')
    without_tenant_id = BodyParameter('withoutTenantId', provide=pycamunda.base.value_is_true)

    def __init__(
        self,
        url: str,
        name: str,
        execution_id: str = None,
        tenant_id: str = None,
        without_tenant_id: bool = False
    ):
        """Deliver a signal to all process definitions and executions with the specific signal
        handler.

        :param url: Camunda Rest engine URL.
        :param name: Name of the signal.
        :param execution_id: Id of the execution to deliver the signal to.
        :param tenant_id: Id of the tenant. Signal will only be delivered to process definitions or
                          executions that belong to this tenant.
        :param without_tenant_id: Whether to deliver the signal only to process definitions or
                                  executions without tenant.
        """
        super().__init__(url=url + URL_SUFFIX)
        self.name = name
        self.execution_id = execution_id
        self.tenant_id = tenant_id
        self.without_tenant_id = without_tenant_id

        self.variables = {}

    def add_variable(
        self,
        name: str,
        value: typing.Any,
        type_: str = None,
        value_info: typing.Mapping = None
    ):
        """Add variables to the process after correlating the message.

        :param name: Name of the variable.
        :param value: Value of the variable.
        :param type_: Value type of the variable.
        :param value_info: Additional information regarding the value type.
        """
        self.variables[name] = {'value': value, 'type': type_, 'valueInfo': value_info}

        return self

    def __call__(self, *args, **kwargs) -> None:
        """Send the request."""
        super().__call__(pycamunda.base.RequestMethod.POST, *args, **kwargs)


class EventAll(_Event):

    def __init__(
        self,
        url: str,
        name: str,
        tenant_id: str = None,
        without_tenant_id: bool = False
    ):
        """Deliver a signal to all process definitions and executions with the specific signal
        handler.

        :param url: Camunda Rest engine URL.
        :param name: Name of the signal.
        :param tenant_id: Id of the tenant. Signal will only be delivered to process definitions or
                          executions that belong to this tenant.
        :param without_tenant_id: Whether to deliver the signal only to process definitions or
                                  executions without tenant.
        """
        super().__init__(
            url=url, name=name, tenant_id=tenant_id, without_tenant_id=without_tenant_id
        )


class EventSingle(_Event):

    def __init__(
        self,
        url: str,
        name: str,
        execution_id: str
    ):
        """Deliver a signal to all process definitions and executions with the specific signal
        handler.

        :param url: Camunda Rest engine URL.
        :param name: Name of the signal.
        :param execution_id: Id of the execution to deliver the signal to.
        """
        super().__init__(url=url, name=name, execution_id=execution_id)
